package com.qipeng.checkstyle.checks;

import antlr.CommonAST;
import com.puppycrawl.tools.checkstyle.api.AbstractCheck;
import com.puppycrawl.tools.checkstyle.api.DetailAST;
import com.puppycrawl.tools.checkstyle.api.TokenTypes;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Spring MVC 注解检查类。
 *
 * @author tieguanyin
 */
@SuppressWarnings("all")
public class SpringMvcAnnotationCheck extends AbstractCheck {

  private static final Map<String, String> REQUEST_MAPPINGS = new HashMap<>();

  static {
    REQUEST_MAPPINGS.put("ALL", "RequestMapping");
    REQUEST_MAPPINGS.put("GET", "GetMapping");
    REQUEST_MAPPINGS.put("POST", "PostMapping");
    REQUEST_MAPPINGS.put("PUT", "PutMapping");
    REQUEST_MAPPINGS.put("PATCH", "PatchMapping");
    REQUEST_MAPPINGS.put("DELETE", "DeleteMapping");
  }

  @Override
  public int[] getDefaultTokens() {
    return getAcceptableTokens();
  }

  @Override
  public int[] getAcceptableTokens() {
    return new int[]{TokenTypes.ANNOTATION};
  }

  @Override
  public int[] getRequiredTokens() {
    return getAcceptableTokens();
  }

  @Override
  public void visitToken(DetailAST ast) {
    DetailAST annotation = ast.findFirstToken(TokenTypes.IDENT);

    // 检查方法上的 @ResponseBody
    if ("RestController".equals(annotation.getText())) {
      DetailAST classAst = ast.getParent().getParent();
      processResponseBodyAnnotation(classAst);
      return;
    }

    // 检查方法上的 @XxxMapping
    if (REQUEST_MAPPINGS.values().contains(annotation.getText())) {
      List<DetailAST> pairs = findAllAnnotationMemberValuePair(ast);
      if (pairs.size() == 1) {
        DetailAST pair = pairs.get(0);
        String paramName = pair.findFirstToken(TokenTypes.IDENT).getText();
        if ("value".equals(paramName) || "path".equals(paramName)) {
          log(pair.getLineNo(), pair.getColumnNo(), "@{0} 的参数名 value/path 可以省略",
            annotation.getText());
        }
        return;
      }

      DetailAST methodAst = ast.getParent().getParent();
      if (methodAst.getType() == TokenTypes.METHOD_DEF) {
        DetailAST methodReturnType = methodAst.findFirstToken(TokenTypes.TYPE)
          .findFirstToken(TokenTypes.IDENT);
        if (methodReturnType != null && "Object".equals(methodReturnType.getText())) {
          log(methodReturnType.getLineNo(), methodReturnType.getColumnNo(),
            "Controller 方法的返回值应该使用确定类型，使用 Object 容易引入 bug");
        }
      }

      if ("RequestMapping".equals(annotation.getText())) {
        String method = pairs.stream()
          .filter(a -> "method".equals(a.findFirstToken(TokenTypes.IDENT).getText()))
          .findFirst()
          .map(a -> a.findFirstToken(TokenTypes.EXPR).getFirstChild())
          .map(a -> a.getChildCount() > 1
            ? a.findFirstToken(TokenTypes.IDENT).getNextSibling() : a)
          .map(CommonAST::getText)
          .orElse(null);
        if (REQUEST_MAPPINGS.containsKey(method)) {
          log(ast.getLineNo(), ast.getColumnNo(), "请将 @{0} 替换为更简洁的 @{1}",
            annotation.getText(), REQUEST_MAPPINGS.get(method));
        }
      }
    }
  }

  private void processResponseBodyAnnotation(DetailAST ast) {
    if (!ast.branchContains(TokenTypes.ANNOTATION)) {
      return;
    }

    DetailAST childAst = ast.getFirstChild();
    while (childAst != null) {
      if (childAst.getType() == TokenTypes.ANNOTATION
        && "ResponseBody".equals(childAst.findFirstToken(TokenTypes.IDENT).getText())) {
        log(childAst.getLineNo(), childAst.getColumnNo(),
          "RestController 方法的 @ResponseBody 可以省略");
      } else {
        processResponseBodyAnnotation(childAst);
      }

      childAst = childAst.getNextSibling();
    }
  }

  private static List<DetailAST> findAllAnnotationMemberValuePair(DetailAST ast) {
    List<DetailAST> asts = new ArrayList<>();
    DetailAST child = ast.getFirstChild();
    while (child != null) {
      if (child.getType() == TokenTypes.ANNOTATION_MEMBER_VALUE_PAIR) {
        asts.add(child);
      }
      child = child.getNextSibling();
    }
    return asts;
  }
}
